/********************************************************************
    
    main.c

    Servo Tester New version

   use rotary encorder, and oled display.

     radicon servo motor controlle signal is changeable
   period time from 5msec to 25msec by 1msec unit, and
   pulse wide from 400usec to 2600usec by 2usec unit.
   pulse wide and period time are displayed in the OLED
   128x64 dot graphic display.

   (c) 2022.7 Suwa-Koubou   http://www.suwa-koubou.jp


    2022.7.9   first release

    pin assign PIC32MX120F032B

                        MCLR 1      28 AVdd  ------- 0.1uF
    Power led --------   RA0 2      27 AVss  --------+
                         RA1 3      26 RB15/IC4  ---  encorder A
         PGED1  ------   RB0 4      25 RB14      ---  encorder B
         PGEC1  ------   RB1 5      24 RB13      ---  encorder Sw
                         RB2 6      23 RB12
                         RB3 7      22 RB11
               GND ----- Vss 8      21 RB10
    20MHz ---+----- OSC1/RA2 9      20 Vcap   ----- 10uF
    Cerarock +----- OSC2/RA3 10     19 Vss    -------+-- GND
                         RB4 11     18 RB9/SDA1 --- OLED display SDA 
                         RA4 12     17 RB8/SCL1 --- OLED display SCL
              +3.3V ---- Vdd 13     16 RB7/OC1  --- Servo motor control signal 
                         RB5 14     15 RB6
    

********************************************************************/

#include <stdio.h>
#include "HardwareProfile.h"
#include "typedef.h"
#include "timer.h"
#include "delay.h"
#include "oled.h"


/*===================================================================
	Configuration for PIC32MX 
===================================================================*/

#include "p32mx120f_ceramic_20_config.h"

/*===================================================================
	definition 
===================================================================*/

#define PULSE_WIDE_MAX   2600
#define PULSE_WIDE_MIN    400
#define PULSE_WIDE_RESET 1500

#define PERIOD_MAX         25
#define PERIOD_MIN          5
#define PERIOD_RESET       20

#define PULSE_MODE          0
#define PERIOD_MODE         1

#define lock_timer2()      IEC0bits.T2IE=0
#define unlock_timer2()    IEC0bits.T2IE=1

#define lock_encorder()    IEC1bits.CNBIE=0
#define unlock_encorder()  IEC1bits.CNBIE=1

/*===================================================================
	global variable 
===================================================================*/

int    pulse_wide;      // 2600 - 400  1500+/-1100  (unit 2usec)
int    period_time;     // 25   - 5    (unit 1msec)

int    encorder_count;
int    mode;

char   pbuf[80];      // print buffer


/*-------------------------------------------------------------------
	function prototype
-------------------------------------------------------------------*/

int   main(void);

void  pulse_control(void);
void  display_time_value(void);
int   check_mode_switch(void);
void  set_pulse_wide(int count);
void  set_period_time(int count);

void  init_sys(void);


/*===================================================================
	main
===================================================================*/

int main(void)
{
    // system initialize
    init_sys();

    pulse_wide = PULSE_WIDE_RESET;
    period_time =PERIOD_RESET;
    mode = PULSE_MODE;
    encorder_count = 0;

    // oled initialize
    oled_init();

    // loop for ever
    pulse_control();

    return 0;
}

/*===================================================================
    pulse control
===================================================================*/

void pulse_control(void)
{
    int  prev_count;
    int  count;


    prev_count = encorder_count;

    for(;;){

        display_time_value();

        if(check_mode_switch()==1){
            oled_clear();
            continue;
        }

        lock_encorder();
        count = encorder_count - prev_count;
        prev_count = encorder_count;
        unlock_encorder();

        if(count == 0) continue;

        if(mode == PULSE_MODE){
            set_pulse_wide(count);
        }else{
            set_period_time(count);
        }    
    }
}


/*-------------------------------------------------------------------
    display pwm period value

    0123456789ABCDEF
    < Servo Tester >
       11550000
    pw:11550000 usec
           2200     
    pr:    2200 msec

-------------------------------------------------------------------*/ 

void display_time_value(void)
{
    oled_locate(0,0); oled_putstr("< Servo Tester >");
    oled_locate(0,4); oled_putstr("< Servo Tester >");

    if(mode == PULSE_MODE){
        //sprintf(pbuf, "%4d", (((pulse_wide*10)/4)*4)/10);
        sprintf(pbuf, "%4d", pulse_wide);
        oled_putnumstr(3,1, pbuf);
        oled_putnumstr(3,5, pbuf);

        oled_locate(0,2); oled_putstr("pw:");
        oled_locate(0,6); oled_putstr("pw:");
        oled_locate(12,2); oled_putstr("usec");
        oled_locate(12,6); oled_putstr("usec");

        sprintf(pbuf, "pr:      %2d msec", period_time);
        oled_locate(0,3); oled_putstr(pbuf);
        oled_locate(0,7); oled_putstr(pbuf);

    }else{
        //sprintf(pbuf, "pw:    %4d usec", (((pulse_wide*10)/4)*4)/10);
        sprintf(pbuf, "pw:    %4d usec", pulse_wide);
        oled_locate(0,1); oled_putstr(pbuf);
        oled_locate(0,5); oled_putstr(pbuf);

        sprintf(pbuf, "%2d", period_time);
        oled_putnumstr(7,2,pbuf);
        oled_putnumstr(7,6,pbuf);

        oled_locate( 0,3); oled_putstr("pr:");
        oled_locate( 0,7); oled_putstr("pr:");
        oled_locate(12,3); oled_putstr("msec");
        oled_locate(12,7); oled_putstr("msec");
    }
}


/*-------------------------------------------------------------------
    check mode switch
      encorder switch connected RB13
-------------------------------------------------------------------*/ 

int check_mode_switch(void)
{
    if(PORTBbits.RB13 == 1){
        delay_ms(50);
        return 0;      // no mode change
    }
       
    delay_ms(20);
    if(PORTBbits.RB13 == 1){
        return 0;
    }

    // mode change
    if(mode == PULSE_MODE){
        mode = PERIOD_MODE;
        LATAbits.LATA0 = 1;   // LED turn off
    }else if(mode == PERIOD_MODE){
        mode = PULSE_MODE;
        LATAbits.LATA0 = 0;   // LED turn on
    }

    // wait for Switch release
    while(PORTBbits.RB13 == 0);
    return 1;
}


/*-------------------------------------------------------------------
    set pulse wide
-------------------------------------------------------------------*/ 

void set_pulse_wide(int count)
{
    int tmp;
    int regval;

    tmp = pulse_wide;

    if(count > 3 || count < -3){
        tmp += (count *10);         // rapid rotary
    }else{
        tmp += (count >0 ? 2:-2);   // slow rotary
                                    // minimum unit 2usec (2/0.4=5clock)
    }

    if(tmp >= PULSE_WIDE_MAX) tmp = PULSE_WIDE_MAX;
    if(tmp <= PULSE_WIDE_MIN) tmp = PULSE_WIDE_MIN;

    regval = (tmp * 10)/4;    // regval = tmp / 0.4
    pulse_wide = tmp;

    OC1RS = regval;
}


/*-------------------------------------------------------------------
    set period time
-------------------------------------------------------------------*/ 

void set_period_time(int count)
{
    lock_timer2();
    if(count > 0 && period_time < PERIOD_MAX){
        period_time++;
    }else if(count < 0 && period_time > PERIOD_MIN){
        period_time--;
    }
    unlock_timer2();
}


/*===================================================================
    encorder CH-A interrupt functions
===================================================================*/

/*-------------------------------------------------------------------
    CH-A RB15 change notice interrupt,  CH-B is RB14

    encorder countup/countdown at the rise edge or fall edge.
             ____      ____      ____
    CH-A ___|    |____|    |____|    |____  RB15 interrupt
               ____      ____      ____
    CH-B _____|    |____|    |____|    |____RB14 level read

            CCW <---      ----> CW
-------------------------------------------------------------------*/ 

void __attribute__ (( interrupt(ipl3auto), vector(_CHANGE_NOTICE_VECTOR) )) _CNBInterrupt(void)
{

    if(PORTBbits.RB15){  // rise edge
        if(PORTBbits.RB14 == 0){
            encorder_count++;
        }else{
            encorder_count--;
        }
    }else{               // fall edge
        if(PORTBbits.RB14 == 0){
            encorder_count--;
        }else{
            encorder_count++;
        }
    }

    IFS1bits.CNBIF = 0;
}


/*===================================================================
    Timer2 interrupt functions
===================================================================*/

/*-------------------------------------------------------------------
    Timer2 interrupt
    update timer2 period time, update timing is every timer timeup.
-------------------------------------------------------------------*/ 

void __attribute__ (( interrupt(ipl2auto), vector(_TIMER_2_VECTOR) )) _T2Interrupt(void)
{
    IFS0bits.T2IF = 0;

    PR2 = period_time * 2500 -1;   // clock count 2500count/1msec
}


/*===================================================================
    system initialize
===================================================================*/ 

void init_sys(void)
{
    __builtin_disable_interrupts();
    
    // Data Memory SRAM wait states: Default Setting = 1; set it to 0
    BMXCONbits.BMXWSDRM = 0;

    // enable multi-vector interrupts
    INTCONSET = _INTCON_MVEC_MASK;
 
    // enable global interrupts
    __builtin_enable_interrupts();


    ANSELA = 0x0000;        // Analog Off
    ANSELB = 0x0000;
    LATA = LATB = 0;        // Low Level Out
    TRISA = 0xffff;         // RA input
    TRISB = 0xffff;         // RB input

    // setup power led
    TRISAbits.TRISA0 = 0;
    LATAbits.LATA0 = 0;    // led on

    // setup encorder CH-A(RB15) change notification
    CNENBbits.CNIEB15 = 1;
    CNCONBbits.ON = 1;
    IPC8bits.CNIP = 3;
    IFS1bits.CNBIF = 0;
    IEC1bits.CNBIE = 1;

    // setup Timer 2
    T2CON = 0;
    T2CONbits.TCKPS = 0b100;   // 1:16  40MHz/16=2.5MHz(0.4usec)
    PR2 = 49999;    // 25msec(25000usec)/0.4usec=62500
                    // 20msec(20000usec)/0.4usec=50000
                    //  5msec( 5000usec)/0.4usec=12500
                    //  1msec( 1000usec)/0.4usec=2500

    // setup OC1
    RPB7R = 0b0101;   // RPB7 --> OC1
    OC1CON = 0;
    OC1R = 3750;      // 2.6msec(2600usec)/0.4usec=6500
                      // 2.0msec(2000usec)/0.4usec=5000
                      // 1.5msec(1500usec)/0.4usec=3750 
                      // 1.0msec(1000usec)/0.4usec=2500
                      // 0.5msec( 500usec)/0.4usec=1250
                      // 0.4msec( 400usec)/0.4usec=1000
                      //            1usec /0.4usec=2.5
                      //            2usec /0.4usec=5 
    OC1RS = 3750;
    OC1CONbits.OCM = 0b110; // PWM

    IPC2bits.T2IP = 2;
    IFS0bits.T2IF = 0;
    IEC0bits.T2IE = 1;

    OC1CONbits.ON = 1;
    T2CONbits.ON = 1;

    // timer initialize
    timer_init();

    // OLED(SSD1306) I2C interface enable
    I2C1CON = 0;
    I2C1BRG = 49;  // PBCLK=40MHz(40000KHz), FSCK=400KHz,
                   // (PBCLK/(2*FSCK))-1=(40000/(2*400))-1=49
    I2C1CONbits.DISSLW = 1;
    I2C1CONbits.I2CEN = 1;

}


/*** end of main **************************************************/
