/*************************************************************************

	File: oled.c


     128x64 dot graphic display driver for SSD1306 chip set .

     2019.6 (C)  S.Suwa <www.suwa-koubou.jp>

     2022/7/8 oled_locate() bug fix.
              update oled_init()


*/

#include  <string.h>
#include  "HardwareProfile.h"
#include  "delay.h"
#include  "oled.h"
//#include  "font.h"   // font size 6
#include  "font8.h"    // font size 8x8
#include  "font16x16.h"  // font size 16x16 digit(0-9) only


/*--------------------------------------------------------------------------
	SSD1306 I2C device definition
--------------------------------------------------------------------------*/ 

#define DEVADDR    (0x3c << 1)    // (0x3c << 1) is 0x78
#define	MW         0     // I2C master write
#define MR         1     // I2C master read

#define DATA_SEND  0x40  // Control Byte Co=0, D/C#=1
#define CMD_SEND   0x00  // Control Byte Co=0, D/C#=0


//#define FONT_SIZE   6     // 5*7  X11 font 
#define FONT_SIZE   8   // 8*8  MZ-80 font


#define SWAP(x,y) { unsigned char d = x; x = y; y = d; }
#define abs(x)    ((x)>0?(x):-(x))


/*------------------------------------------------------------------------
    global variable
------------------------------------------------------------------------*/

static int font_size;

/*------------------------------------------------------------------------
    function prototype
------------------------------------------------------------------------*/

void  oled_init(void);
void  oled_clear(void);
void  oled_locate(int col, int row);
void  oled_putstr(char *str);
void  oled_putnumstr(int col, int row, char *str);
void  oled_putch(int c);

static void  oled_write_data(char *data, int len);
static void  oled_write_cmd(char *cmd, int len);


// I2C handling
static void  i2cOpenStart(void);
//static void  i2cReStart(void); // do not use
static void  i2cStopClose(void);
static int   i2cSend(char data);

#if 0
static char  i2cRec(void);
static void  i2cAckSend(void);
static void  i2cNAckSend(void);
#endif

/*=========================================================================
	oled_initialize
=========================================================================*/

void oled_init(void)
{
    font_size = FONT_SIZE;

    delay_ms(100);

    // 0xAE      Display Off
    // 0xA8 0x3F Set MULTIPLEX (64line) 64dot
    // 0xC0      COM output scan direction 0 to 63
    // 0xDA 0x02 Set COM Pins Hardware Configuration/alternative configuration
    // 0x8D 0x14 Set charge pump enable 
    // 0x81 0xFF Set Contrast Control/ max contrast
    // 0xAF      Display ON
    // other default value

    oled_write_cmd("\xAE\xA8\x3F\xC0\xDA\x02\x8D\x14\x81\xFF\xAF", 11);
    oled_clear();
}


/*=========================================================================
	oled_clear)()
=========================================================================*/

void oled_clear(void)
{
    int n;

    // 0x20 Set memory addressing mode
    // 0x00 Horizontal addressing mode
    // 0x21 Set column address
    // 0x00 column start address =0
    // 0x7f column end address   =127
    // 0x22 Set page address
    // 0x00 page start address =0
    // 0x03 page end address   =7
    oled_write_cmd("\x20\x00\x21\x00\x7f\x22\x00\x07", 8);

    // sending clear data
	i2cOpenStart();
	i2cSend(DEVADDR +MW);
	i2cSend(DATA_SEND);
    for(n=0; n<1024; n++){ // 128column*8page
	    i2cSend(0x00);
    }
	i2cStopClose();

    oled_locate(0,0);

}


/*=========================================================================
	oled_locate()

    2022/7/8 bug fix.  '& 0xff' --> '& 0x0f'

    col: 0-20 (max 21column) at the font_size = 6
         0-15 (max 16column) at the font_size = 8
         0-7  (max  8column) at the font_size = 16
    row: 0-7  (max 8row)
=========================================================================*/

void oled_locate(int col, int row)
{
    char data[5];

    data[0]=0x20;  // Set memory addressing mode
    data[1]=0x02;  // Page addressing mode
    data[2]=0xB0 + row; // Set page start address for Page addressing mode
    //data[3]= (col*font_size) & 0xff;  // Set lower column start address for Page addressing mode
    data[3]= (col*font_size) & 0x0f;  // Set lower column start address for Page addressing mode
    data[4]= 0x10 + ((col*font_size) >> 4);  // Set higher column start address for Page addressing mode

    oled_write_cmd(data, 5);
}


/*=========================================================================
	oled_putstr()
    can't automatic write to next line at reached End Column.
=========================================================================*/

void oled_putstr(char *str)
{
    char data[8];

    while(*str){
        if(*str < ' ' || *str > 0xff){
            memset(data, 0, font_size);
        }else{
            memcpy(data, font[*str -0x20], font_size);
        }
        oled_write_data(data, font_size);
        str++;
    }
}

/*=========================================================================
	oled_putnumstr()   using 16x16 font, set locate position by 8x8font
=========================================================================*/

void oled_putnumstr(int col, int row, char *str)
{
    char data[16];
    char *num;
    int  n;

    oled_locate(col, row);
    num = str;
    while(*num){
        if(*num < '0' || *num > '9'){
            memset(data, 0, 16);
        }else{
            for(n=0; n<16; n++){
                data[n] = font16x16[*num -'0'][n] & 0xff;  // low byte
            }
        }
        oled_write_data(data, 16);
        num++;
    }

    oled_locate(col, row+1);
    num = str;
    while(*num){
        if(*num < '0' || *num > '9'){
            memset(data, 0, 16);
        }else{
            for(n=0; n<16; n++){
                data[n] = (font16x16[*num -'0'][n] >> 8) & 0xff;  // high byte
            }
        }
        oled_write_data(data, 16);
        num++;
    }
}


/*=========================================================================
	oled_putch()
    can't automatic write to next line at reached End Column.
=========================================================================*/

void oled_putch(int c)
{
    char data[8];

    if(c < ' ' || c > 0xff){
        memset(data, 0, font_size);
    }else{
        memcpy(data, font[c -0x20], font_size);
    }
    oled_write_data(data, font_size);
}

/*--------------------------------------------------------------------------
	oled_write_data()
--------------------------------------------------------------------------*/
static
void oled_write_data(char *data, int len)
{
	i2cOpenStart();
	i2cSend(DEVADDR +MW);
	i2cSend(DATA_SEND);
    while(len--){
	    i2cSend(*data++);
    }
	i2cStopClose();
	//delay_ms(1);
}	


/*--------------------------------------------------------------------------
	oled_write_cmd()
--------------------------------------------------------------------------*/
static
void oled_write_cmd(char *cmd, int len)
{
	i2cOpenStart();
	i2cSend(DEVADDR +MW);
	i2cSend(CMD_SEND);
    while(len--){
	    i2cSend(*cmd++);
    }
	i2cStopClose();
	//delay_ms(1);
}	



/*==========================================================================
	I2C Handling module
	Test Board used I2C1
==========================================================================*/
#define	TIME_OUT	0xffff
static	unsigned short tw;

#define I2CnCONbits  I2C1CONbits
//#define I2CnCONbits  I2C2CONbits
//#define I2CnCONbits  I2C3CONbits

#define I2CnSTATbits  I2C1STATbits
//#define I2CnSTATbits  I2C2STATbits
//#define I2CnSTATbits  I2C3STATbits

#define I2CnTRN    I2C1TRN
//#define I2CnTRN    I2C2TRN
//#define I2CnTRN    I2C3TRN

#define I2CnRCV    I2C1RCV
//#define I2CnRCV    I2C2RCV
//#define I2CnRCV    I2C3RCV


/*--------------------------------------------------------------------------
	I2C master Open and Start
--------------------------------------------------------------------------*/
static
void	i2cOpenStart(void)
{
	// enabled I2C
	//I2CnCONbits.I2CEN = 1;

	// START 
	I2CnCONbits.SEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CnCONbits.SEN) break;
	}
}


/*--------------------------------------------------------------------------
	I2C master ReStart
--------------------------------------------------------------------------*/
#if 0
static
void	i2cReStart(void)
{
	// RESTART 
	I2CnCONbits.RSEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CnCONbits.RSEN) break;
	}
}
#endif

/*--------------------------------------------------------------------------
	I2C master Stop and Close
--------------------------------------------------------------------------*/

static
void	i2cStopClose(void)
{
	// STOP 
	I2CnCONbits.PEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CnCONbits.PEN)break;
	}

	// disabled I2C
	//I2CnCONbits.I2CEN = 0;
}


/*--------------------------------------------------------------------------
	I2C master data send
--------------------------------------------------------------------------*/

static
int	i2cSend(char data)
{
	I2CnTRN = data;

    if(I2CnSTATbits.IWCOL){	// If write collision occurs,return -2
		return -2;
    }

	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CnSTATbits.TRSTAT) break;  // wait until write cycle is complete 
	}

	if (I2CnSTATbits.ACKSTAT){ 	// test for ACK condition received
		return -1;	// NACK received
	}

	return 0;	// complete with no error.
}


/*--------------------------------------------------------------------------
	I2C	master data receive
--------------------------------------------------------------------------*/
#if 0
static
char	i2cRec(void)
{
	// enabled Receive
	I2CnCONbits.RCEN = 1;

	// wait until receive complete and data stored to I2CRCV register
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CnCONbits.RCEN && I2CnSTATbits.RBF) break;
	}

	// clear overflow bit
	I2CnSTATbits.I2COV = 0;
    
	return (char)I2CnRCV;
}


/*--------------------------------------------------------------------------
	I2C master Ack send
--------------------------------------------------------------------------*/

static
void	i2cAckSend(void)
{
	// Ack Send
	I2CnCONbits.ACKDT = 0;
	I2CnCONbits.ACKEN = 1;

	// wait until ack cycle complete
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CnCONbits.ACKEN)break;
	}
}


/*--------------------------------------------------------------------------
	I2C master NotAck send
--------------------------------------------------------------------------*/

static
void	i2cNAckSend(void)
{
	// NAck Send
	I2CnCONbits.ACKDT = 1;
	I2CnCONbits.ACKEN = 1;

	// wait until ack cycle complete
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CnCONbits.ACKEN) break;
	}
}

#endif  

/*** end of oled.c ******************************************************/
